SuperStrict

Import "MatrixVector.bmx"
Import "FastPerlinNoise.bmx"
Include "Voxelmap Generator.bmx"
Include "Texture.bmx"
Include "Entity.bmx"
Include "Camera.bmx"
Include "Light.bmx"
Include "VBO.bmx"
Include "Mesh.bmx"
Include "Chunk.bmx"
Include "Renderer.bmx"
Include "LuaHandler.bmx"
Include "BlockClass.bmx"

TRenderer.Graphics3D( 1024, 768 )

TLuaHandler.Init()
TBlockClass.Init()

Global Camera:TCamera = TCamera.Create()
Camera.SetAOV( 60.0 )
Camera.SetViewport( 0, 0, 1024, 768 )
Camera.SetClsColor( 128, 128, 128 )
Camera.SetRange( 1.0, 500.0 )
Camera.Move( 0.0, 100.0, 0.0 )

TLight.SetAmbientColor( 180, 180, 160 )

Local Light:TLight = TLight.Create( TLight.KIND_DIRECTIONAL )
Light.Rotate( 45.0, -30.0, 0.0 )
Light.SetColor( 255, 255, 128 )

Global OldChunkX:Int, OldChunkZ:Int, ViewRange:Int = 5, ChunksToDo:Int
Global Pitch:Int, Yaw:Int = 90

Local Timer:TTimer = CreateTimer( 60 )

Init()

While Not ( KeyHit( KEY_ESCAPE ) Or AppTerminate() )
	TRenderer.RenderWorld()
	
	CreateVisibleChunks()
	
	TChunk.UpdateChunks()
	
	UserInput()
	
	Flip 0
	WaitTimer Timer
Wend
End

Function Init()
	For Local X:Int = ViewRange*2 - 1 To 0 Step -1
		For Local Z:Int = 0 Until ViewRange*2
			TChunk.CreateChunk( X - ViewRange, Z - ViewRange, ChunkCallback )
			
			ChunksToDo :+ 1
		Next
	Next
	
	MoveMouse TRenderer.GWidth/2, TRenderer.GHeight/2
	MouseXSpeed()
	MouseYSpeed()
End Function

Function UserInput()
	Pitch :- MouseYSpeed()
	Yaw   :+ MouseXSpeed()
	
	Pitch = Min( Max( -90.0, Pitch ), 90.0 )
	
	Camera.Rotate( Pitch, 0.0, 0.0 )
	Camera.Turn( 0.0, Yaw, 0.0 )
	
	MoveMouse TRenderer.GWidth/2, TRenderer.GHeight/2
	MouseXSpeed()
	MouseYSpeed()
	
	If ChunksToDo < ViewRange Then Camera.Move( 0.2, 0.0, 0.0, True )
	
	'Local Speed:Float = 0.2
	'Camera.Move( ( KeyDown( KEY_RIGHT ) - KeyDown( KEY_LEFT ) )*Speed, ( KeyDown( KEY_SPACE ) - KeyDown( KEY_LCONTROL ) )*Speed, ( KeyDown( KEY_DOWN ) - KeyDown( KEY_UP ) )*Speed )
End Function

Function CreateVisibleChunks()
	Local CamChunkX:Int = Camera.Position.X/TChunk.CHUNK_WIDTH
	Local CamChunkZ:Int = Camera.Position.Z/TChunk.CHUNK_WIDTH
	
	Local DiffX:Int = OldChunkX - CamChunkX
	Local DiffZ:Int = OldChunkZ - CamChunkZ
	
	If DiffX Or DiffZ Then
		For Local Chunk:TChunk = EachIn TRenderer.Meshes
			If Abs( Chunk.ChunkX - CamChunkX ) > ViewRange*2 Or Abs( Chunk.ChunkZ - CamChunkZ ) > ViewRange*2 Then TRenderer.Meshes.Remove( Chunk )
		Next
		
		For Local X:Int = 0 Until Abs( DiffX )
			For Local Z:Int = 0 Until ViewRange*2
				TChunk.CreateChunk( -Sgn( DiffX )*( ViewRange + X ) + OldChunkX, CamChunkZ - ViewRange + Z, ChunkCallback )
			Next
		Next
		
		For Local Z:Int = 0 Until Abs( DiffZ )
			For Local X:Int = 0 Until ViewRange*2
				TChunk.CreateChunk( CamChunkX - ViewRange + X, -Sgn( DiffZ )*( ViewRange + Z ) + OldChunkZ, ChunkCallback )
			Next
		Next
	EndIf
	
	OldChunkX = CamChunkX
	OldChunkZ = CamChunkZ
End Function

Function ChunkCallback( Chunk:TChunk )
	ChunksToDo :- 1
End Function